from pathlib import Path
import re
import sys
import traceback
import click # Import click

# Relative imports for package modules
# from .parsing import parse_text_to_json, format_json_to_solidity_struct
from .format import detect_and_load_input, format_user_op_data, format_to_user_op_json # Updated imports
from .hashing import calculate_user_op_hash, eip191_hash_hex, eip191_hash_message, hex_to_bytes
from .signature import recover_signer
from .debug import encode_debug_command
from .simulate import encode_simulate_command
from .constants import ENTRY_POINT_V07

DEFAULT_ENTRY_POINT = ENTRY_POINT_V07

def get_version():
    init = Path("src/kek/__init__.py").read_text()
    return re.search(r'^__version__ = ["\']([^"\']+)["\']', init, re.M).group(1)

# --- Main Click Group --- 
@click.group(context_settings=dict(help_option_names=['-h', '--help']))
@click.version_option(package_name='zerodev-kek', message='%(package)s version %(version)s') # Added standard version option
def cli():
    """KEK: Kernel Examination Kit

    A CLI tool to parse, format, hash, verify signatures, and debug 
    EIP-4337 UserOperations, especially for Kernel.
    """
    pass

# --- `version` subcommand ---
@cli.command('version')
def version_cmd():
    """Display the installed version of KEK."""
    version = get_version()
    click.echo(f"kek version {version}")

# --- Helper to load input (used by multiple commands) ---
def load_input_data(raw_input_str):
    try:
        user_op_intermediate_data = detect_and_load_input(raw_input_str)
        if not user_op_intermediate_data:
            click.echo("\nError: Could not parse input data.", err=True)
            sys.exit(1)
        return user_op_intermediate_data
    except Exception as e:
        click.echo(f"\nError processing input: {e}", err=True)
        traceback.print_exc() # Still useful for debugging
        sys.exit(1)

# --- `format` command ---
@cli.command('format')
@click.argument('raw_input', type=str)
@click.option('--output', '-o', type=click.Choice(['packed', 'userop'], case_sensitive=False), default='packed', 
              help="Output format: 'packed' (default) or 'userop'.")
def format_cmd(raw_input, output):
    """Parse/normalize input and output JSON in specified format."""
    user_op_intermediate_data = load_input_data(raw_input)
    
    try:
        if output == 'userop':
            output_json = format_to_user_op_json(user_op_intermediate_data)
            click.echo("--- Standard UserOperation JSON ---")
            click.echo(output_json)
        else: # Default is 'packed'
            output_json = format_user_op_data(user_op_intermediate_data)
            click.echo("--- Formatted PackedUserOperation JSON ---")
            click.echo(output_json)
    except Exception as e:
        click.echo(f"\nAn error occurred during formatting: {e}", err=True)
        traceback.print_exc()
        sys.exit(1)

# --- `userOpHash` command ---
@cli.command('userOpHash')
@click.argument('raw_input', type=str)
@click.option('--chainId', '-c', type=int, required=True, help="Chain ID for hash calculation.")
@click.option('--entrypoint', '-e', default=DEFAULT_ENTRY_POINT, 
              help=f"EntryPoint address (default: {DEFAULT_ENTRY_POINT}).")
def user_op_hash_cmd(raw_input, chainid, entrypoint):
    """Calculate the EIP-4337 UserOperation hash."""
    user_op_intermediate_data = load_input_data(raw_input)
    try:
        final_json = format_user_op_data(user_op_intermediate_data)
        user_op_hash = calculate_user_op_hash(final_json, entrypoint, chainid)
        click.echo("--- Calculated UserOpHash ---")
        click.echo(user_op_hash)
    except Exception as e:
        click.echo(f"\nAn error occurred during hash calculation: {e}", err=True)
        traceback.print_exc()
        sys.exit(1)

# --- `signer` command ---
@cli.command('signer')
@click.argument('raw_input', type=str)
@click.option('--chainId', '-c', type=int, required=True, help="Chain ID for hash calculation.")
@click.option('--entrypoint', '-e', default=DEFAULT_ENTRY_POINT, 
              help=f"EntryPoint address (default: {DEFAULT_ENTRY_POINT}).")
@click.option('--signer', '-s', 'expected_signer_address', # Store in this variable
              help="Optional: Verify signature against this address. If flag used without address, show all results.")
@click.option('--verify-only', is_flag=True, default=False, help="Only verify against --signer, don't show all results if flag used alone.")
def signer_cmd(raw_input, chainid, entrypoint, expected_signer_address, verify_only):
    """Recover signer or verify signature against expected address."""
    user_op_intermediate_data = load_input_data(raw_input)
    
    # Determine mode based on presence and type of expected_signer_address
    mode = 'skip'
    if expected_signer_address is not None:
        # Validate address if provided
        if isinstance(expected_signer_address, str):
             if not expected_signer_address.startswith('0x') or len(expected_signer_address) != 42:
                 click.echo(f"Error: Invalid format for --signer address: {expected_signer_address}", err=True)
                 sys.exit(1)
             try: bytes.fromhex(expected_signer_address[2:])
             except ValueError: 
                 click.echo(f"Error: Invalid hex characters in --signer address: {expected_signer_address}", err=True)
                 sys.exit(1)
             mode = 'verify'
        else: # Flag used alone (const=True not directly usable here, check type if needed or rely on click validation)
             if not verify_only:
                 mode = 'show_all'
             else:
                 # If --verify-only is set but no address given, it's an error or just skip
                 click.echo("Error: --verify-only requires an address provided via --signer.", err=True)
                 sys.exit(1)

    if mode == 'skip':
        click.echo("\n--- Signature Recovery Skipped (no --signer flag provided) ---")
        return

    try:
        # Format and calculate hashes needed for recovery
        final_json = format_user_op_data(user_op_intermediate_data)
        user_op_hash = calculate_user_op_hash(final_json, entrypoint, chainid)
        user_op_hash_bytes = hex_to_bytes(user_op_hash)
        eip191_hash_of_hash_bytes_hex = eip191_hash_hex(user_op_hash)
        eip191_digest_bytes = hex_to_bytes(eip191_hash_of_hash_bytes_hex)
        eip191_hash_of_hash_string_hex = eip191_hash_message(user_op_hash)
        eip191_digest_string_bytes = hex_to_bytes(eip191_hash_of_hash_string_hex)

        click.echo("\n--- Signature Recovery --- ")
        signature_hex = user_op_intermediate_data.get("signature")
        if not signature_hex:
             for k, v in user_op_intermediate_data.items():
                 if k.lower() == "signature": signature_hex = v; break

        if not signature_hex or signature_hex == '0x':
            click.echo("No signature provided in input.")
            return
        elif len(signature_hex) != 132:
            click.echo(f"Invalid signature length: {len(signature_hex)} chars (expected 132)", err=True)
            return
        
        # Perform recovery
        recovered_from_userophash = recover_signer(user_op_hash_bytes, signature_hex)
        recovered_from_eip191_bytes = recover_signer(eip191_digest_bytes, signature_hex)
        recovered_from_eip191_string = recover_signer(eip191_digest_string_bytes, signature_hex)

        if mode == 'verify':
            click.echo(f"Verifying signature against expected signer: {expected_signer_address}")
            click.echo(f"Signature provided: {signature_hex}")
            match_found_signer = False
            if recovered_from_userophash and recovered_from_userophash.lower() == expected_signer_address.lower():
                click.echo("  ✅ Matches signer for Digest 1 (UserOpHash Bytes)")
                match_found_signer = True
            if recovered_from_eip191_bytes and recovered_from_eip191_bytes.lower() == expected_signer_address.lower():
                click.echo("  ✅ Matches signer for Digest 2 (EIP-191 Bytes)")
                match_found_signer = True
            if recovered_from_eip191_string and recovered_from_eip191_string.lower() == expected_signer_address.lower():
                click.echo("  ✅ Matches signer for Digest 3 (EIP-191 String)")
                match_found_signer = True
            if not match_found_signer:
                click.echo(f"  ❌ Signature did NOT recover specified signer ({expected_signer_address}).")
        
        elif mode == 'show_all':
            sender_from_op = user_op_intermediate_data.get("sender")
            if not sender_from_op: 
                for k, v in user_op_intermediate_data.items():
                    if k.lower() == "sender": sender_from_op = v; break
            if sender_from_op: click.echo(f"Sender (from Op): {sender_from_op}")
            click.echo(f"Signature:        {signature_hex}")
            click.echo("\nShowing all recovery results:")
            click.echo("-"*20)
            click.echo(f"Digest 1 (UserOpHash): 0x{user_op_hash_bytes.hex()}")
            click.echo(f"  Recovered: {recovered_from_userophash or 'Failed'}")
            click.echo("-"*20)
            click.echo(f"Digest 2 (EIP-191 Bytes): 0x{eip191_digest_bytes.hex()}")
            click.echo(f"  Recovered: {recovered_from_eip191_bytes or 'Failed'}")
            click.echo("-"*20)
            click.echo(f"Digest 3 (EIP-191 String): 0x{eip191_digest_string_bytes.hex()}")
            click.echo(f"  Recovered: {recovered_from_eip191_string or 'Failed'}")
            click.echo("-"*20)

    except Exception as e:
        click.echo(f"\nAn error occurred during signer recovery/verification: {e}", err=True)
        traceback.print_exc()
        sys.exit(1)

# --- `debug` command ---
@cli.command('debug')
@click.argument('raw_input', type=str)
@click.option('--rpc-url', required=True, help="RPC URL for the cast command.")
def debug_cmd(raw_input, rpc_url):
    """Generate or execute `cast call --trace` for EntryPoint.handleOps."""
    user_op_intermediate_data = load_input_data(raw_input)
    # Pass args as a simple object/dict if needed by encode_debug_command
    class Args: pass
    debug_args = Args()
    debug_args.rpc_url = rpc_url
    
    encode_debug_command(debug_args, user_op_intermediate_data)

# --- `simulate` command ---
@cli.command('simulate')
@click.argument('raw_input', type=str)
@click.option('--rpc-url', required=True, help="RPC URL for the estimation command.")
def simulate_cmd(raw_input, rpc_url):
    """Simulate a UserOperation with Pimlico estimation address."""
    user_op_intermediate_data = load_input_data(raw_input)
    # Pass args as a simple object/dict if needed by encode_simulate_command
    class Args: pass
    simulate_args = Args()
    simulate_args.rpc_url = rpc_url
    
    try:
        encode_simulate_command(simulate_args, user_op_intermediate_data)
    except Exception as e:
        click.echo(f"\nAn error occurred during simulation: {e}", err=True)
        traceback.print_exc()
        sys.exit(1)

# --- Entry point for `kek` command ---
if __name__ == '__main__':
    cli() 


from eth_keys.exceptions import BadSignature
from eth_keys import keys

# Use relative import for shared utility
from .utils import hex_to_bytes

def recover_signer(digest_bytes: bytes, signature_hex: str) -> str | None:
    """Attempts to recover the signer's address from a digest and signature.

    Args:
        digest_bytes: The 32-byte hash (digest) that was signed.
        signature_hex: The signature as a 0x-prefixed hex string (65 bytes).

    Returns:
        The checksummed signer address as a string, or None if recovery fails.
    """
    if not signature_hex or signature_hex == '0x' or len(signature_hex) != 132:
        return None
    try:
        signature_bytes = hex_to_bytes(signature_hex)
        # if v is 27 or 28, set v to 0
        if signature_bytes[64] == 27 or signature_bytes[64] == 28:
            signature_bytes = signature_bytes[:64] + b'\x00'
        
        if len(digest_bytes) != 32:
            # print(f"Debug: Digest must be 32 bytes for recovery, got {len(digest_bytes)}.")
            return None

        signature = keys.Signature(signature_bytes=signature_bytes)
        public_key = signature.recover_public_key_from_msg_hash(digest_bytes)
        return public_key.to_checksum_address()
    except (BadSignature, ValueError, TypeError, Exception): # Catch potential errors
        # print(f"Debug: Recovery failed for digest {digest_bytes.hex()} - {e}")
        return None 


import sys
from eth_abi import encode
import json

# Use relative imports
from .utils import hex_to_bytes
# from .parsing import format_json_to_solidity_struct # Need to format first
from .format import format_user_op_data # Updated import
from .constants import ENTRY_POINT_V07

def encode_debug_command(args, user_op_intermediate_data):
    """Generates and optionally executes the handleOps cast call command."""
    # print("\n--- Debug handleOps Call --- ") # Reduced verbosity
    beneficiary = "0x0000000000000000000000000000000000000000" # Hardcoded zero address

    # -- Prepare UserOperation data for encoding -- 
    try:
        # 1. Format the intermediate data dictionary into the final PackedUserOp JSON string
        final_json_string = format_user_op_data(user_op_intermediate_data)
        # 2. Load the final formatted JSON string into a dictionary for tuple creation
        user_op_dict = json.loads(final_json_string) 
    except ValueError as e:
        print(f"Error formatting/loading UserOperation data: {e}")
        sys.exit(1)
    except Exception as e: # Catch other potential errors
        print(f"Unexpected error during UserOperation formatting/loading: {e}")
        sys.exit(1)

    try:
        user_op_tuple = (
            user_op_dict['sender'],
            int(user_op_dict['nonce']),
            hex_to_bytes(user_op_dict['initCode']),
            hex_to_bytes(user_op_dict['callData']),
            hex_to_bytes(user_op_dict['accountGasLimits']),
            int(user_op_dict['preVerificationGas']),
            hex_to_bytes(user_op_dict['gasFees']),
            hex_to_bytes(user_op_dict['paymasterAndData']),
            hex_to_bytes(user_op_dict['signature'])
        )
    except KeyError as e:
         print(f"Error: Missing key '{e}' in formatted UserOperation JSON needed for ABI encoding.")
         sys.exit(1)
    except (ValueError, TypeError) as e:
         print(f"Error converting UserOperation fields for ABI encoding: {e}")
         sys.exit(1)
         
    ops_array = [user_op_tuple]

    user_op_abi_type = '(address,uint256,bytes,bytes,bytes32,uint256,bytes32,bytes,bytes)'
    handle_ops_input_types = [f'{user_op_abi_type}[]', 'address']
    
    # -- ABI Encode handleOps call --
    try:
        encoded_call_data = encode(
            handle_ops_input_types,
            [ops_array, beneficiary] # Use hardcoded zero address
        )
    except Exception as e:
        print(f"\nError ABI encoding handleOps data: {e}")
        sys.exit(1)

    handle_ops_selector = "0x765e827f" # v0.7 selector
    full_calldata = handle_ops_selector + encoded_call_data.hex()
    
    # -- Construct cast command list and string --
    cast_command_list = [
        "cast", "call", 
        ENTRY_POINT_V07, 
        full_calldata, 
        "--rpc-url", args.rpc_url, 
        "--trace"
    ]
    cast_command_string = ' '.join(cast_command_list) # For printing

    print(cast_command_string)



import json

# Use relative imports for shared utility

# --- Parsing Helpers ---

def parse_gas_value_to_wei(value) -> int:
    """Parses various gas value inputs (str, int, float) into wei (int)."""
    if isinstance(value, int):
        return value # Assume already in wei
    if isinstance(value, float):
        return int(value * 1e9) # Assume float represents gwei
    
    if not isinstance(value, str):
        raise ValueError(f"Cannot parse gas value of type {type(value)}")

    value_str = value.strip().lower()
    
    try:
        if value_str.endswith(" gwei"):
            num_part = value_str.replace(" gwei", "").strip()
            return int(float(num_part) * 1e9)
        elif value_str.startswith("0x"):
            return int(value_str, 16) # Assume hex is wei
        elif '.' in value_str or 'e' in value_str:
            return int(float(value_str) * 1e9) # Assume float notation is gwei
        else:
            return int(value_str) # Assume numeric string is wei
    except ValueError as e:
        raise ValueError(f"Could not parse gas value '{value}' to wei: {e}")

def to_int_if_hex(value):
    """Converts a hex string (0x...) to int, otherwise returns original."""
    if isinstance(value, str) and value.startswith('0x'):
        try:
            return int(value, 16)
        except ValueError:
            return value # Return original if not valid hex
    return value

def unpack_uint128_pair(packed_hex: str) -> tuple[int, int]:
    """Unpacks a bytes32 hex string into two uint128 integers."""
    if not isinstance(packed_hex, str) or not packed_hex.startswith('0x') or len(packed_hex) != 66:
        raise ValueError(f"Invalid packed hex format for unpacking: '{packed_hex}'")
    try:
        val1_hex = packed_hex[2:34]  # First 128 bits (32 hex chars)
        val2_hex = packed_hex[34:]   # Second 128 bits (32 hex chars)
        return int(val1_hex, 16), int(val2_hex, 16)
    except ValueError as e:
        raise ValueError(f"Could not unpack hex string '{packed_hex}': {e}")

# --- Input Handling & Normalization ---

def detect_and_load_input(input_str: str) -> dict:
    """Detects input format and normalizes to a consistent intermediate dict format."""
    input_str = input_str.strip()
    is_likely_json = input_str.startswith('{') and input_str.endswith('}')
    data = {}

    if is_likely_json:
        try:
            raw_data = json.loads(input_str)
            normalized_data = {k.lower(): v for k, v in raw_data.items()} 
            data.update(normalized_data)
            # print("Detected JSON input.")
            # Check if it's likely PackedUserOp JSON AFTER initial load
            if all(k in data for k in ["sender", "nonce", "accountgaslimits", "gasfees"]):
                # print("Detected PackedUserOperation JSON input. Unpacking...")
                # Unpack gas limits
                try:
                    vgl_unpacked, cgl_unpacked = unpack_uint128_pair(data.get('accountgaslimits', '0x' + '0'*64))
                    data['verificationgaslimit'] = vgl_unpacked
                    data['callgaslimit'] = cgl_unpacked
                except ValueError as e:
                    print(f"Warning: Failed to unpack accountGasLimits from input: {e}")
                # Unpack gas fees
                try:
                    prio_unpacked, max_unpacked = unpack_uint128_pair(data.get('gasfees', '0x' + '0'*64))
                    # Note the order: prio first, then max in Packed format
                    data['maxpriorityfeepergas'] = prio_unpacked
                    data['maxfeepergas'] = max_unpacked
                except ValueError as e:
                     print(f"Warning: Failed to unpack gasFees from input: {e}")
                # We don't delete the packed keys, normalization below will handle types
            
            # Fallthrough to normalize all loaded JSON data

        except json.JSONDecodeError:
            is_likely_json = False # Treat as raw text if JSON parsing fails

    if not is_likely_json:
        # print("Assuming raw text input.")
        lines = input_str.splitlines()
        temp_data = {}
        for line in lines:
            line = line.strip()
            if not line: continue
            colon_index = line.find(':')
            if colon_index == -1: continue
            key = line[:colon_index].strip().lower() # Normalize key
            value_str = line[colon_index + 1:].strip()
            temp_data[key] = value_str # Store value as string for now
        data.update(temp_data)

    # --- Normalize values to consistent types (applies to ALL input types now) --- 
    normalized_output = {}
    int_keys = ["nonce", "callgaslimit", "verificationgaslimit", "preverificationgas", "paymasterverificationgaslimit", "paymasterpostopgaslimit"]
    gas_keys = ["maxfeepergas", "maxpriorityfeepergas"]
    # Include packed keys here, but they will be handled as hex strings
    hex_keys = ["sender", "initcode", "calldata", "accountgaslimits", "gasfees", "paymasteranddata", "signature", "factory", "factorydata", "paymaster", "paymasterdata"]

    for key, value in data.items():
        l_key = key.lower()
        output_key = l_key 
        
        if output_key in int_keys:
            try:
                # Ensure it's not None before converting
                if value is None: raise TypeError("None value for int key")
                normalized_output[output_key] = int(value, 16) if isinstance(value, str) and value.startswith('0x') else int(value)
            except (ValueError, TypeError):
                print(f"Warning: Could not convert '{key}' value '{value}' to int. Setting to 0.")
                normalized_output[output_key] = 0 # Default to 0 on error
        elif output_key in gas_keys:
             try:
                if value is None: raise TypeError("None value for gas key")
                normalized_output[output_key] = parse_gas_value_to_wei(value)
             except (ValueError, TypeError) as e:
                 print(f"Warning: Could not parse gas value for '{key}': {e}. Setting to 0.")
                 normalized_output[output_key] = 0
        elif output_key in hex_keys:
             if isinstance(value, str):
                 if value.startswith('0x'):
                     normalized_output[output_key] = value
                 elif not value.startswith('0x') and all(c in '0123456789abcdefABCDEF' for c in value):
                     normalized_output[output_key] = '0x' + value
                 elif value == "": # Handle empty strings for bytes
                      normalized_output[output_key] = '0x'
                 else:
                     # Might be a non-hex string mistakenly assigned (e.g. from raw text)
                     print(f"Warning: Non-hex value '{value}' found for expected hex field '{key}'. Setting to '0x'.")
                     normalized_output[output_key] = '0x' 
             elif value is None:
                 normalized_output[output_key] = '0x'
             else:
                 print(f"Warning: Unexpected type '{type(value)}' for hex field '{key}'. Setting to '0x'.")
                 normalized_output[output_key] = '0x'
        else:
             normalized_output[output_key] = value
             
    return normalized_output


def format_user_op_data(user_op_data: dict) -> str:
    """
    Formats normalized UserOperation data (dict) into the final 
    PackedUserOperation JSON string, handling optional fields and packing.
    """
    # Helper to safely get data (case-insensitive keys already handled by normalization)
    def get_data(key, default=None):
        return user_op_data.get(key, default)

    # --- Optional Field Defaults & Constants --- 
    ZERO_ADDRESS = '0x' + '0' * 40
    EMPTY_BYTES = '0x'
    ZERO_BYTES32 = '0x' + '0' * 64

    # --- Format/Pack Fields --- 

    # Helper to format uint128 to hex (32 chars, no prefix)
    def format_uint128_hex_noprefix(val):
         val = to_int_if_hex(val) # Convert hex string to int first if needed
         if not isinstance(val, int):
            try: val = int(val)
            except (ValueError, TypeError): raise ValueError(f"Cannot convert '{val}' to int for uint128")
         if val < 0 or val >= (1 << 128):
             raise ValueError(f"Value {val} out of range for uint128")
         return f"{val:032x}"

    # Get core fields (expecting specific types from normalization)
    sender = get_data('sender', ZERO_ADDRESS)
    nonce = get_data('nonce', 0)
    preVerificationGas = get_data('preverificationgas', 0)
    callData = get_data('calldata', EMPTY_BYTES)
    signature = get_data('signature', EMPTY_BYTES)
    maxFeePerGas = get_data('maxfeepergas', 0)
    maxPriorityFeePerGas = get_data('maxpriorityfeepergas', 0)
    callGasLimit = get_data('callgaslimit', 0)
    verificationGasLimit = get_data('verificationgaslimit', 0)
    
    # Handle initCode (prioritize existing, then construct)
    initCode = get_data('initcode', None)
    if initCode is None:
        factory = get_data('factory')
        factoryData = get_data('factorydata', EMPTY_BYTES) 
        if factory and factory != ZERO_ADDRESS:
            f_data_hex = factoryData[2:] if factoryData.startswith('0x') else factoryData
            initCode = f"{factory}{f_data_hex}"
        else:
            initCode = EMPTY_BYTES
    elif not isinstance(initCode, str) or not initCode.startswith('0x'):
         print(f"Warning: Provided initCode '{initCode}' is not a valid hex string. Using '0x'.")
         initCode = EMPTY_BYTES

    # Handle paymasterAndData (prioritize existing, then construct)
    paymasterAndData = get_data('paymasteranddata', None)
    if paymasterAndData is None:
        paymaster = get_data('paymaster')
        paymasterData = get_data('paymasterdata', EMPTY_BYTES)
        if paymaster and paymaster != ZERO_ADDRESS:
            pm_data_hex = paymasterData[2:] if paymasterData.startswith('0x') else paymasterData
            try:
                pmVg = get_data('paymasterverificationgaslimit', 0)
                pmPg = get_data('paymasterpostopgaslimit', 0)
                pm_ver_gas_hex = format_uint128_hex_noprefix(pmVg)
                pm_post_gas_hex = format_uint128_hex_noprefix(pmPg)
                paymasterAndData = f"{paymaster}{pm_ver_gas_hex}{pm_post_gas_hex}{pm_data_hex}"
            except ValueError as e:
                print(f"Warning: Could not format paymaster gas limits ({e}). Omitting gas limits.")
                paymasterAndData = f"{paymaster}{pm_data_hex}" # Fallback without gas limits
        else:
            paymasterAndData = EMPTY_BYTES
    elif not isinstance(paymasterAndData, str) or not paymasterAndData.startswith('0x'):
         print(f"Warning: Provided paymasterAndData '{paymasterAndData}' is not a valid hex string. Using '0x'.")
         paymasterAndData = EMPTY_BYTES

    # Pack accountGasLimits
    try:
        accountGasLimits = f"0x{format_uint128_hex_noprefix(verificationGasLimit)}{format_uint128_hex_noprefix(callGasLimit)}"
    except ValueError as e:
        print(f"Warning: Could not format accountGasLimits ({e}). Defaulting.")
        accountGasLimits = ZERO_BYTES32

    # Pack gasFees
    try:
        gasFees = f"0x{format_uint128_hex_noprefix(maxPriorityFeePerGas)}{format_uint128_hex_noprefix(maxFeePerGas)}"
    except ValueError as e:
         print(f"Warning: Could not format gasFees ({e}). Input Values: Prio={maxPriorityFeePerGas}, Max={maxFeePerGas}. Defaulting.")
         gasFees = ZERO_BYTES32

    # Final Output Dict Construction
    output_dict = {
        "sender": sender,
        "nonce": str(nonce), # Store as string
        "initCode": initCode,
        "callData": callData,
        "accountGasLimits": accountGasLimits,
        "preVerificationGas": str(preVerificationGas), # Store as string
        "gasFees": gasFees,
        "paymasterAndData": paymasterAndData,
        "signature": signature
    }

    return json.dumps(output_dict, indent=2)

def format_to_user_op_json(user_op_data: dict) -> str:
    """
    Formats normalized UserOperation data (dict) into the standard
    EIP-4337 UserOperation JSON structure (numbers as hex strings).
    Handles optional fields, using defaults if missing.
    """
    # Helper to safely get data (case-insensitive keys already handled by normalization)
    def get_data(key, default=None):
        return user_op_data.get(key, default)

    # --- Optional Field Defaults & Constants ---
    ZERO_ADDRESS = '0x' + '0' * 40
    EMPTY_BYTES = '0x'
    ZERO_INT = 0

    # --- Get Normalized Data (expecting ints/hex strings from detect_and_load_input) ---
    # Use lowercase keys as defined in the normalization step
    sender = get_data('sender', ZERO_ADDRESS)
    nonce = get_data('nonce', ZERO_INT)
    factory = get_data('factory') # Get potential factory address
    factoryData = get_data('factorydata') # Get potential factory data
    initCode = get_data('initcode') # Get potential pre-built initCode
    callData = get_data('calldata', EMPTY_BYTES)
    callGasLimit = get_data('callgaslimit', ZERO_INT)
    verificationGasLimit = get_data('verificationgaslimit', ZERO_INT)
    preVerificationGas = get_data('preverificationgas', ZERO_INT)
    maxFeePerGas = get_data('maxfeepergas', ZERO_INT) # Expecting wei int
    maxPriorityFeePerGas = get_data('maxpriorityfeepergas', ZERO_INT) # Expecting wei int
    paymaster = get_data('paymaster') # Get potential paymaster address
    paymasterVerificationGasLimit = get_data('paymasterverificationgaslimit', ZERO_INT) # Use ZERO_INT default
    paymasterPostOpGasLimit = get_data('paymasterpostopgaslimit', ZERO_INT) # Use ZERO_INT default
    paymasterData = get_data('paymasterdata') # Get potential paymaster data
    paymasterAndData = get_data('paymasteranddata') # Get potential pre-built paymasterAndData
    signature = get_data('signature', EMPTY_BYTES)

    # --- Build the UserOperation JSON dictionary IN ORDER --- 
    # Using a temporary list of tuples to maintain order before json.dumps
    ordered_items = []

    # Add sender and nonce first
    ordered_items.append(("sender", sender))
    ordered_items.append(("nonce", hex(to_int_if_hex(nonce))))

    # Handle Factory/InitCode
    if factory and factory != ZERO_ADDRESS:
        ordered_items.append(("factory", factory))
        ordered_items.append(("factoryData", factoryData if factoryData else EMPTY_BYTES))
    elif initCode and initCode != EMPTY_BYTES:
        # initCode = bytes20(factory) + factoryData
        # make sure you don't forget the 0x prefix, and initCode is also prefixed with 0x
        ordered_items.append(("factory", initCode[:42]))
        ordered_items.append(("factoryData", "0x" + initCode[42:]))
        # print("Warning: Including pre-built 'initCode'. Cannot reliably split into factory/factoryData.")
    # else: implicitly skip if neither factory nor initCode is present

    # Add core execution fields
    ordered_items.append(("callData", callData if callData else EMPTY_BYTES))
    ordered_items.append(("callGasLimit", hex(to_int_if_hex(callGasLimit))))
    ordered_items.append(("verificationGasLimit", hex(to_int_if_hex(verificationGasLimit))))
    ordered_items.append(("preVerificationGas", hex(to_int_if_hex(preVerificationGas))))
    ordered_items.append(("maxFeePerGas", hex(maxFeePerGas)))
    ordered_items.append(("maxPriorityFeePerGas", hex(maxPriorityFeePerGas)))

    # Handle Paymaster/PaymasterAndData
    if paymaster and paymaster != ZERO_ADDRESS:
        ordered_items.append(("paymaster", paymaster))
        ordered_items.append(("paymasterVerificationGasLimit", hex(to_int_if_hex(paymasterVerificationGasLimit))))
        ordered_items.append(("paymasterPostOpGasLimit", hex(to_int_if_hex(paymasterPostOpGasLimit))))
        ordered_items.append(("paymasterData", paymasterData if paymasterData else EMPTY_BYTES))
    elif paymasterAndData and paymasterAndData != EMPTY_BYTES:
        # paymasterAndData = bytes20(paymaster) + paymasterVerificationGasLimit + paymasterPostOpGasLimit + paymasterData
        # make sure you don't forget the 0x prefix, and paymasterAndData is also prefixed with 0x
        ordered_items.append(("paymaster", paymasterAndData[:42]))
        ordered_items.append(("paymasterVerificationGasLimit", hex(to_int_if_hex("0x" + paymasterAndData[42:74])))) # format this to uint hex
        ordered_items.append(("paymasterPostOpGasLimit", hex(to_int_if_hex("0x" + paymasterAndData[74:106])))) # format this to uint hex
        ordered_items.append(("paymasterData", "0x" + paymasterAndData[106:]))
    # else: implicitly skip if neither paymaster info nor paymasterAndData is present

    # Add signature last
    ordered_items.append(("signature", signature if signature else EMPTY_BYTES))

    # Convert the ordered list of tuples into a proper OrderedDict and use standard json serialization
    # Python 3.7+ dictionaries preserve insertion order, so this will maintain our key order
    ordered_dict = {}
    for key, value in ordered_items:
        ordered_dict[key] = value
    
    # Use standard json.dumps with proper indentation
    return json.dumps(ordered_dict, indent=2)

# Remove original parse_text_to_json function if it exists above
# Remove original format_json_to_solidity_struct function if it exists above 


from .constants import PIMLICO_ESTIMATION_ADDRESS, ENTRY_POINT_V07, SIMULATE_ENTRY_POINT_SELECTOR, SIMULATE_HANDLE_OP_LAST_SELECTOR
from eth_abi import encode, decode
import sys
import json
from web3 import Web3

from .format import format_user_op_data
from .utils import hex_to_bytes, decode_simulate_lastOp_error, decode_simulate_lastOp_result, to_cast_trace_command

SIMULATE_TARGET = "0xf384fddcaf70336dca46404d809153a0029a0253"
def encode_simulate_command(args, user_op_intermediate_data):
    """Generate a command to simulate a UserOperation using Pimlico's estimation contract."""
    
    # -- Prepare UserOperation data for encoding -- 
    try:
        # 1. Format the intermediate data dictionary into the final PackedUserOp JSON string
        final_json_string = format_user_op_data(user_op_intermediate_data)
        # 2. Load the final formatted JSON string into a dictionary for tuple creation
        user_op_dict = json.loads(final_json_string) 
    except ValueError as e:
        print(f"Error formatting/loading UserOperation data: {e}")
        sys.exit(1)
    except Exception as e: # Catch other potential errors
        print(f"Unexpected error during UserOperation formatting/loading: {e}")
        sys.exit(1)

    try:
        user_op_tuple = (
            user_op_dict['sender'],
            int(user_op_dict['nonce']),
            hex_to_bytes(user_op_dict['initCode']),
            hex_to_bytes(user_op_dict['callData']),
            hex_to_bytes(user_op_dict['accountGasLimits']),
            int(user_op_dict['preVerificationGas']),
            hex_to_bytes(user_op_dict['gasFees']),
            hex_to_bytes(user_op_dict['paymasterAndData']),
            hex_to_bytes(user_op_dict['signature'])
        )
    except KeyError as e:
         print(f"Error: Missing key '{e}' in formatted UserOperation JSON needed for ABI encoding.")
         sys.exit(1)
    except (ValueError, TypeError) as e:
         print(f"Error converting UserOperation fields for ABI encoding: {e}")
         sys.exit(1)

    simulation_result = {}
    # simulate the user op last
    try:
        simulation_result.execution_result = simulate_user_op_last(args.rpc_url, user_op_tuple)
    except FailedUserOp as e:
        print(e)
        cast_command_list = e.command()
        cast_command_list.append("--rpc-url")
        cast_command_list.append(args.rpc_url)
        cast_command_string = ' '.join(cast_command_list)
        print(cast_command_string)
        sys.exit(1)

    # validation_result = decode( ['((uint256,uint256,uint256,uint256,bytes),(uint256,uint256),(uint256,uint256),(uint256,uint256),(address,(uint256,uint256)))'], decoded_results[1])
    # print(validation_result)
    # -- Construct cast command list and string --
    # cast_command_list = [
    #     "cast", "call", 
    #     PIMLICO_ESTIMATION_ADDRESS, 
    #     simulate_calldata, 
    #     "--rpc-url", args.rpc_url,
    #     "--trace"
    # ]
    # cast_command_string = ' '.join(cast_command_list)  # For printing

    # print(cast_command_string)

def simulate_user_op_last(rpc_url,user_op_tuple) -> object:
    ops_array = [user_op_tuple]

    # --- Step 1: Encode the simulateHandleOpLast call ---
    user_op_abi_type = '(address,uint256,bytes,bytes,bytes32,uint256,bytes32,bytes,bytes)'
    
    try:
        # Encode simulateHandleOpLast with just the user ops
        encoded_simulate_last = encode(
            [f'{user_op_abi_type}[]'],  # Just the user ops array
            [ops_array]
        )
    except Exception as e:
        print(f"\nError encoding simulateHandleOpLast data: {e}")
        sys.exit(1)

    simulate_last_data = SIMULATE_HANDLE_OP_LAST_SELECTOR + encoded_simulate_last.hex()

    # --- Step 2: Encode the simulateEntryPoint call ---
    try:
        # Encode simulateEntryPoint with the entrypoint and simulateHandleOpLast calldata
        encoded_simulate = encode(
            ["address", "bytes[]"],
            [ENTRY_POINT_V07, [bytes.fromhex(simulate_last_data[2:])]]  # Remove 0x prefix
        )
    except Exception as e:
        print(f"\nError encoding simulateEntryPoint parameters: {e}")
        sys.exit(1)
        
    simulate_calldata = SIMULATE_ENTRY_POINT_SELECTOR + encoded_simulate.hex()
    # check simulation return data
    w3 = Web3(Web3.HTTPProvider(rpc_url))
    abi = """
[
    {
        "inputs": [{"internalType": "address", "name": "entryPoint", "type": "address"}, {"internalType": "bytes[]", "name": "userOps", "type": "bytes[]"}],
        "name": "simulateEntryPoint",
        "outputs": [{"internalType": "bytes[]", "name": "results", "type": "bytes[]"}],
        "stateMutability": "view",
        "type": "function"
    }
    ]
    """

    """
    struct ReturnInfo {
        uint256 preOpGas;
        uint256 prefund;
        uint256 accountValidationData;
        uint256 paymasterValidationData;
        bytes paymasterContext;
    }

    struct AggregatorStakeInfo {
        address aggregator;
        StakeInfo stakeInfo;
    }

    struct StakeInfo {
        uint256 stake;
        uint256 unstakeDelaySec;
    }

    struct ValidationResult {
        ReturnInfo returnInfo;
        StakeInfo senderInfo;
        StakeInfo factoryInfo;
        StakeInfo paymasterInfo;
        AggregatorStakeInfo aggregatorInfo;
    }
    """

    """
    struct ExecutionResult {
        uint256 preOpGas;
        uint256 paid;
        uint256 accountValidationData;
        uint256 paymasterValidationData;
        uint256 paymasterVerificationGasLimit;
        uint256 paymasterPostOpGasLimit;
        bool targetSuccess;
        bytes targetResult;
    }
    """
    contract = w3.eth.contract(address=Web3.to_checksum_address(PIMLICO_ESTIMATION_ADDRESS), abi=abi)
    results = contract.functions.simulateEntryPoint(ENTRY_POINT_V07, [bytes.fromhex(simulate_last_data[2:])]).call()
    decoded_results = decode( ['bool', 'bytes'], results[0][4:])
    # if simulation was successful, print the execution result
    if decoded_results[0]:
        details = decode_simulate_lastOp_result(decoded_results[1])
        return details
    else:
        raise FailedUserOp(decode_simulate_lastOp_error(decoded_results[1]), simulate_calldata)


class FailedUserOp(Exception):
    def __init__(self, message, debugCallData):
        self.message = message
        self.debugCallData = debugCallData
        super().__init__(self.message)

    def __str__(self):
        return f"Failed UserOp : {self.message[:4]}"
    
    def command(self):
        return to_cast_trace_command(SIMULATE_TARGET, self.debugCallData)


from eth_abi import decode
def hex_to_bytes(hex_string: str) -> bytes:
    """Converts a hex string (with or without 0x prefix) to bytes."""
    if not isinstance(hex_string, str):
         raise TypeError(f"Input must be a string, got {type(hex_string)}")
    if hex_string.startswith('0x'):
        # Handle empty bytes case
        if len(hex_string) == 2:
            return b''
        return bytes.fromhex(hex_string[2:])
    # Handle empty bytes case for non-prefixed string
    if not hex_string:
        return b''
    try:
        return bytes.fromhex(hex_string)
    except ValueError as e:
        raise ValueError(f"Invalid hex string for hex_to_bytes: '{hex_string[:20]}...'") from e 

def to_cast_trace_command(target_address, call_data):
    return ["cast", "call", target_address, call_data, "--trace"]

def decode_simulate_lastOp_error(error_bytes: bytes) -> str:
    """Decodes an error bytes into a string."""
    # FailedOp(uint256,string) == 0x220266b6
    # 00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000040000000000000000000000000000000000000000000000000000000000000001a4141323520696e76616c6964206163636f756e74206e6f6e6365000000000000
    if error_bytes.hex().startswith("220266b6"):
        return decode(['uint256', 'string'], error_bytes[4:])[1]
    else:
        return "Unknown error"
    
def decode_simulate_lastOp_result(result_bytes: bytes) -> str:
    """Decodes a result bytes into a string."""
    decoded = decode(['(uint256,uint256,uint256,uint256,uint256,uint256,bool,bytes)'], result_bytes)
    # formatted execution result
    formatted_execution_result = {
        "preOpGas": decoded[0][0],
        "paid": decoded[0][1],
        "accountValidationData": decoded[0][2],
        "paymasterValidationData": decoded[0][3],
        "paymasterVerificationGasLimit": decoded[0][4],
        "paymasterPostOpGasLimit": decoded[0][5],
        "targetSuccess": decoded[0][6],
        "targetResult": decoded[0][7]
    }
    return formatted_execution_result   


ENTRY_POINT_V07 = "0x0000000071727De22E5E9d8BAf0edAc6f37da032"
PIMLICO_ESTIMATION_ADDRESS = "0x74cb5e4ee81b86e70f9045036a1c5477de69ee87"

# Correct function selectors
SIMULATE_ENTRY_POINT_SELECTOR = "0xc18f5226"
SIMULATE_HANDLE_OP_LAST_SELECTOR = "0x263934db"  # simulateHandleOpLast selector




import re
from pathlib import Path
from setuptools import setup, find_packages

def get_version():
    content = Path("src/kek/__init__.py").read_text()
    return re.search(r'^__version__\s*=\s*["\']([^"\']+)["\']', content, re.M).group(1)

setup(
    name="zerodev-kek",
    version=get_version(),
    description="Kernel Examination Kit - A CLI tool for EIP-4337",
    author="taek lee",
    author_email="leekt216@gmail.com",
    url="https://github.com/zerodevapp/kek",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    entry_points={
        "console_scripts": ["kek = kek.cli:cli"]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)



import json
# Removed: import sha3
from eth_abi import encode
from eth_utils import keccak as eth_keccak # Renamed import

# Use relative import for shared utility
# from utils import hex_to_bytes 
from .utils import hex_to_bytes 
# Import the unpacking function

def calculate_user_op_hash(user_op_json: str, entry_point_address: str, chain_id: int) -> str:
    """Calculates the userOpHash according to EIP-4337 EntryPoint logic.
    
    Expects user_op_json to be the output of format_user_op_data, containing
    pre-packed fields like accountGasLimits and gasFees.

    Args:
        user_op_json: JSON string of the PackedUserOperation.
        entry_point_address: The address of the EntryPoint contract.
        chain_id: The chain ID.

    Returns:
        The userOpHash as a '0x'-prefixed hex string.
    """
    try:
        user_op = json.loads(user_op_json)
    except json.JSONDecodeError as e:
        raise ValueError(f"Invalid UserOperation JSON provided: {e}")

    # --- Extract values and hash byte fields --- 
    try:
        sender = user_op['sender']
        nonce = int(user_op['nonce']) # Stored as string, convert to int
        pre_verification_gas = int(user_op['preVerificationGas']) # Stored as string, convert to int

        # These should be bytes32 hex strings from formatting step
        account_gas_limits_bytes = hex_to_bytes(user_op['accountGasLimits']) 
        gas_fees_bytes = hex_to_bytes(user_op['gasFees'])
        if len(account_gas_limits_bytes) != 32:
            raise ValueError(f"accountGasLimits must be 32 bytes hex, got {len(account_gas_limits_bytes)} bytes from {user_op['accountGasLimits']}")
        if len(gas_fees_bytes) != 32:
             raise ValueError(f"gasFees must be 32 bytes hex, got {len(gas_fees_bytes)} bytes from {user_op['gasFees']}")

        # These are bytes fields from formatting step, hash them now
        init_code_bytes = hex_to_bytes(user_op['initCode']) 
        call_data_bytes = hex_to_bytes(user_op['callData'])
        paymaster_and_data_bytes = hex_to_bytes(user_op['paymasterAndData'])
        
        hash_init_code = eth_keccak(init_code_bytes)
        hash_call_data = eth_keccak(call_data_bytes)
        hash_paymaster_and_data = eth_keccak(paymaster_and_data_bytes)

        # Validate entry point address format
        if not (isinstance(entry_point_address, str) and entry_point_address.startswith('0x') and len(entry_point_address) == 42):
            raise ValueError("EntryPoint address must be a 0x-prefixed hex string of 42 chars")
            
    except KeyError as e:
        raise ValueError(f"Missing required key in formatted UserOperation JSON: {e}")
    except (ValueError, TypeError) as e:
        raise ValueError(f"Invalid format in formatted UserOperation JSON or arguments: {e}")

    # --- ABI Encode and Hash --- 
    packed_user_op_part_for_hash = encode(
        ['address', 'uint256', 'bytes32', 'bytes32', 'bytes32', 'uint256', 'bytes32', 'bytes32'],
        [sender, nonce, hash_init_code, hash_call_data, account_gas_limits_bytes, pre_verification_gas, gas_fees_bytes, hash_paymaster_and_data]
    )
    user_op_partial_hash = eth_keccak(packed_user_op_part_for_hash)
    packed_final_for_hash = encode(
        ['bytes32', 'address', 'uint256'],
        [user_op_partial_hash, entry_point_address, chain_id]
    )
    user_op_hash = eth_keccak(packed_final_for_hash)
    return '0x' + user_op_hash.hex()

# --- EIP-191 Hashing Functions --- 

def eip191_hash_bytes(data_bytes: bytes) -> str:
    """Calculates the EIP-191 hash for the given bytes."""
    if len(data_bytes) != 32:
         print(f"Warning: EIP-191 hashing typically expects 32-byte input for hashes, got {len(data_bytes)}.")
    prefix = b'\x19Ethereum Signed Message:\n' + str(len(data_bytes)).encode('ascii')
    message_to_hash = prefix + data_bytes
    # Replaced pysha3 with eth_utils.keccak
    # hasher = sha3.keccak_256()
    # hasher.update(message_to_hash)
    # return '0x' + hasher.hexdigest()
    hash_bytes = eth_keccak(message_to_hash)
    return '0x' + hash_bytes.hex()

def eip191_hash_message(message: str) -> str:
    """Calculates the EIP-191 hash for a UTF-8 string message."""
    try:
        message_bytes = message.encode('utf-8')
    except AttributeError:
         raise TypeError(f"Input message must be a string, got {type(message)}")
    prefix = b'\x19Ethereum Signed Message:\n' + str(len(message_bytes)).encode('ascii')
    message_to_hash = prefix + message_bytes
    # Replaced pysha3 with eth_utils.keccak
    # hasher = sha3.keccak_256()
    # hasher.update(message_to_hash)
    # return '0x' + hasher.hexdigest()
    hash_bytes = eth_keccak(message_to_hash)
    return '0x' + hash_bytes.hex()

def eip191_hash_hex(hex_data: str) -> str:
    """Calculates the EIP-191 hash for data provided as a hex string."""
    try:
        data_bytes = hex_to_bytes(hex_data)
    except (ValueError, TypeError) as e:
        raise ValueError(f"Invalid hex string for EIP-191 hashing: {e}")
    return eip191_hash_bytes(data_bytes) 